/* Main header for Cygnus

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#define _GNU_SOURCE

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <grp.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include "config.h"
#include "sysconf.h"
#include "response.h"

/* True or False? This is used for toggles mainly. */
typedef enum {ERROR = -1,FALSE, TRUE} Boolean_T;

/* Flags */
#define RUN_LIVE	0x00000001	/* Don't fork		*/
#define RUN_SHUTDOWN	0x00000002	/* Shut down		*/
#define RUN_RESTART	0x00000004	/* Shut down, restart	*/
#define RUN_STARTING	0x00000008	/* Starting up		*/
#define RUN_SAVEDBS	0x00000010	/* Save the DBs		*/

/* User Flags */
#define UF_SRA		0x00000001	/* Is an SRA		*/
#define UF_ENFORCED	0x00000002	/* Enforced, waiting	*/

/* NickName Flags */
#define NF_ENFORCE	0x00000001	/* NickName Enforcement	*/
#define NF_SECURE	0x00000002	/* NickName Security	*/
#define NF_HIDEMAIL	0x00000004	/* Hide E-Mail		*/
#define NF_SHOWMAIL	0x00000008	/* Show E-Mail		*/
#define NF_NEVEROP	0x00000010	/* Never Op (ChanServ)	*/
#define NF_NOOP		0x00000020	/* Can't add to CS Lists*/
#define NF_NOSUCCESSOR	0x00000040	/* Can't be successor	*/
#define NF_PRIVATE	0x00000080	/* Privacy Settings	*/
#define NF_NOTICE	0x00000100	/* Use NOTICE		*/
#define NF_PRIVMSG	0x00000200	/* Use PRIVMSG		*/
#define NF_NOMEMO	0x00000400	/* No Memos		*/
#define NF_RECEIPTS	0x00000800	/* Memo Receipts	*/
#define NF_MEMOMAIL	0x00001000	/* MemoMail		*/
#define NF_FROZEN	0x00002000	/* NickName is frozen	*/
#define NF_HELD		0x00004000	/* NickName is held	*/
#define NF_MARKED	0x00008000	/* NickName is marked	*/

/* Special NickName Flags */
#define NF_LINKED	0x10000000	/* Linked NickName	*/
#define NF_CSOP		0x20000000	/* Services Operator	*/
#define NF_NEWS		0x40000000	/* Hasn't read the NEWS	*/
#define NF_WAITAUTH	0x80000000	/* Waiting for Auth	*/

/* NickName Status Flags */
#define NS_RECOGNIZED	0x00000001	/* User is recognized	*/
#define NS_ENFORCED	0x00000002	/* Is an enforcer	*/

/* ChanServ Flags */
#define CF_VERBOSE	0x00000001	/* Verbose reporting	*/
#define CF_VOPALL	0x00000002	/* Auto +v on join	*/
#define CF_SECURE	0x00000004	/* Channel Security	*/
#define CF_RESTRICTED	0x00000008	/* Restricted Channel	*/
#define CF_FROZEN	0x00000010	/* Channel is frozen	*/
#define CF_HELD		0x00000020	/* Channel is held	*/
#define CF_MARKED	0x00000040	/* Channel is marked	*/

/* Special Channel Flags */
#define CF_WAITAUTH	0x10000000	/* Waiting for Auth	*/
#define CF_WAITING	0x20000000	/* Waiting for Verify	*/

/* Memo Status Flags */
#define MF_UNREAD	0x00000001	/* Memo is Unread	*/
#define MF_EMAILED	0x00000002	/* Memo was E-Mailed	*/

/* User Modes */
#define UMODE_o		0x00000001	/* Opered		*/
#define UMODE_r		0x00000002	/* Identified		*/
#define UMODE_A		0x00000004	/* Server Admin		*/
#define UMODE_a		0x00000008	/* Services Admin	*/

#ifdef UNREAL3
#define UMODE_N		0x00000008	/* Network Admin	*/
#define UMODE_T		0x00000010	/* Technical Admin	*/
#endif

/* Channel Modes */
#define CMODE_t		0x00000001	/* Only ops change topic*/
#define CMODE_n		0x00000002	/* No external messages	*/
#define CMODE_s		0x00000004	/* Secret Channel	*/
#define CMODE_m		0x00000008	/* Moderated		*/
#define CMODE_l		0x00000010	/* Limited users	*/	
#define CMODE_i		0x00000020	/* Invite only		*/
#define CMODE_p		0x00000040	/* Private		*/
#define CMODE_k		0x00000080	/* Keyed		*/
#define CMODE_o		0x00000100	/* Op			*/
#define CMODE_v		0x00000200	/* Voice		*/

#ifndef IRCUP9
#define CMODE_R		0x00000400	/* Registered users only*/
#define CMODE_r		0x00000800	/* Registered channel	*/
#endif

#if defined (PROMETHEUS) || defined (BAHAMUT) || defined (UNREAL3)
#define CMODE_c		0x00001000	/* No colors allowed	*/
#define CMODE_O		0x00002000	/* Oper Only channel	*/
#endif

#ifdef UNREAL3
#define CMODE_Q		0x00004000	/* No non-U:Lined kicks	*/
#define CMODE_S		0x00008000	/* Strip color codes	*/
#define CMODE_K		0x00010000	/* No knocking		*/
#define CMODE_V		0x00020000	/* No inviting		*/
#define CMODE_f		0x00040000	/* Flood trigger	*/
#define CMODE_H		0x00080000	/* No hidden users	*/
#define CMODE_G		0x00100000	/* Strip bad words	*/
#define CMODE_C		0x00200000	/* No channel CTCPs	*/
#define CMODE_u		0x00400000	/* Auditorium		*/
#define CMODE_z		0x00800000	/* Secure users only	*/
#define CMODE_N		0x01000000	/* No nick changes	*/
#define CMODE_L		0x02000000	/* Linked Channel	*/
#define CMODE_A		0x04000000	/* Admin Only channel	*/
#define CMODE_h		0x08000000	/* Half-Op		*/
#define CMODE_a		0x10000000	/* Protected		*/
#define CMODE_q		0x20000000	/* Channel Founder	*/
#endif

/* NickServ Timer Stuff */
#define TO_COLLIDE	0		/* Collide this nick?	*/
#define TO_RELEASE	1		/* Release an enforcer?	*/
#define TO_MESS40	2		/* 20 seconds left	*/
#define TO_MESS20	3		/* 40 seconds left	*/

/* ChanServ Timer Stuff */
#define TO_UNINHABIT	0		/* Part the channel	*/

/* Mode stacking stuff */
#define MAXMODES	6		/* Maximum stacked modes*/
#define MAXPARAMSLEN	(510-NICKLEN-CHANNELLEN-34-(7+MAXMODES))

/* Channel hash */
#define CHASH(chan) ((chan)[1] ? ((chan)[1]&31)<<5 | ((chan)[2]&31) : 0)

/* Nickname hash */
#define NHASH(nick) (((nick)[0]&31)<<5 | ((nick)[1]&31))

/* ChanInfo hash */
#define CSHASH(chan) (hashtable[(unsigned char)((chan)[1])]<<5 | \
		      (chan[1] ? hashtable[(unsigned char)((chan)[2])] : 0))

/* NickInfo hash */
#define NSHASH(nick) (hashtable[(unsigned char)((nick)[0])]<<5 | \
		      hashtable[(unsigned char)((nick)[1])])

/* Verification hash */
#define VHASH(name) (((name)[0]&31)<<5 | ((name)[1]&31))

/* E-Mail hash */
#define EHASH(email) (((email)[0]&31)<<5 | ((email)[1]&31))

/* AKill hash */
#define AHASH(akill) (((akill)[0]&31)<<5 | ((akill)[1]&31))

/* Clone hash */
#define CLHASH(host) (((host)[0]&31)<<5 | ((host)[1]&31))

/* Hash table for nicks and channels */
static const char hashtable[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

     0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
    16,17,18,19,20,21,22,23,24,25,26,27,28,29, 0, 0,
     0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
    16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,30,

    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,

    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
};

/* Structure for our stuff */
struct me
{
    char uplink[HOSTLEN];	/* Address we connect to	*/
    int port;			/* Port we do it on		*/
    char pass[PASSWDLEN];	/* Password we use for the link	*/
    char ruplink[HOSTLEN];	/* Real name of our uplink	*/
    char bindhost[HOSTLEN];	/* Address to bind to		*/
    int bindport;		/* Port to bind to		*/
    int token;			/* True if our uplink has tokens*/
    char name[HOSTLEN];		/* Our server name		*/
    char desc[REALLEN];		/* Our description		*/
    time_t since;		/* When we started		*/
    time_t reset;		/* When we last reset		*/
} me;

/* RootServ's information */
struct rs
{
    char nick[NICKLEN];		/* Nick we want to have		*/
    char user[USERLEN];		/* RootServ's username		*/
    char host[HOSTLEN];		/* RootServ's hostname		*/
    char mode[12];		/* RootServ's modes		*/
    char real[REALLEN];		/* RootServ's realname		*/
    int isonline;		/* Is RootServ online?		*/
    int wantnick;		/* Are we waiting for our nick?	*/
    int collided;		/* How many collides?		*/
    int version;		/* Version of RootServ's DB	*/
    time_t lastcoll;		/* Time of last collide		*/
} rs;

/* NickServ's information */
struct ns
{
    char nick[NICKLEN];		/* Nick we want to have		*/
    char user[USERLEN];		/* NickServ's username		*/
    char host[HOSTLEN];		/* NickServ's hostname		*/
    char mode[12];		/* NickServ's modes		*/
    char real[REALLEN];		/* NickServ's realname		*/
    int isonline;		/* Is NickServ online?		*/
    int wantnick;		/* Are we waiting for our nick?	*/
    int collided;		/* How many collides?		*/
    int version;		/* Version of NickServ's DB	*/
    time_t lastcoll;		/* Time of last collide		*/
} ns;

/* ChanServ's information */
struct cs
{
    char nick[NICKLEN];		/* Nick we want to have		*/
    char user[USERLEN];		/* ChanServ's username		*/
    char host[HOSTLEN];		/* ChanServ's hostname		*/
    char mode[12];		/* ChanServ's modes		*/
    char real[REALLEN];		/* ChanServ's realname		*/
    int isonline;		/* Is ChanServ online?		*/
    int wantnick;		/* Are we waiting for our nick?	*/
    int collided;		/* How many collides?		*/
    int version;		/* Version of ChanServ's DB	*/
    time_t lastcoll;		/* Time of last collide		*/
} cs;

/* MemoServ's information */
struct ms
{
    char nick[NICKLEN];		/* Nick we want to have		*/
    char user[USERLEN];		/* MemoServ's username		*/
    char host[HOSTLEN];		/* MemoServ's hostname		*/
    char mode[12];		/* MemoServ's modes		*/
    char real[REALLEN];		/* MemoServ's realname		*/
    int isonline;		/* Is MemoServ online?		*/
    int wantnick;		/* Are we waiting for our nick?	*/
    int collided;		/* How many collides?		*/
    time_t lastcoll;		/* Time of last collide		*/
} ms;

/* GlobalNoticer's information */
struct gn
{
    char nick[NICKLEN];		/* Nick we want to have		*/
    char user[USERLEN];		/* GlobalNoticer's username	*/
    char host[HOSTLEN];		/* GlobalNoticer's hostname	*/
    char mode[12];		/* GlobalNoticer's modes	*/
    char real[REALLEN];		/* GlobalNoticer's realname	*/
    int isonline;		/* Is the GlobalNoticer online? */
    int wantnick;		/* Are we waiting for our nick?	*/
    int collided;		/* How many collides?		*/
    time_t lastcoll;		/* Time of last collide		*/
} gn;

typedef struct user_ User;
typedef struct channel_ Channel;
typedef struct sra_ SRA;
typedef struct akill_ AutoKill;
typedef struct verify_ Verify;
typedef struct email_ EMail;
typedef struct timeout_ Timeout;
typedef struct nickinfo_ NickInfo;
typedef struct chaninfo_ ChanInfo;
typedef struct memo_ Memo;
typedef struct clone_ Clone;
typedef struct trigger_ Trigger;
typedef struct exception_ Exception;

/* User structure: One per user */
struct user_ {
    User *next, *prev;
    char nick[NICKLEN];		/* NickName			*/
    char *user;			/* UserName			*/
    char *host;			/* HostName			*/
#ifdef UNREAL3
    char *fakehost;		/* Fake (masked) hostname	*/
#endif
    char *real;			/* RealName			*/
    char *server;		/* Server			*/
    char *lastnick;		/* Last nickname user id'd for	*/
    char **idchans;		/* Channels user id'd for	*/
    int32 idchancnt;		/* Number of channels id'd for	*/
    int32 flags;		/* User Flags			*/
    int32 mode;			/* UserModes			*/
    int32 sstamp;		/* Services Stamp		*/
    int32 passfail;		/* # of failed IDENTIFY's	*/
    int32 msgs;			/* Number of messages		*/
    int32 offences;		/* How many floods		*/
    time_t lastmsg;		/* Time of last message		*/
    time_t timestamp;		/* TimeStamp			*/
    time_t lastnreg;		/* Time of last nick REGISTER	*/
    time_t lastcreg;		/* Time of last chan REGISTER	*/
    time_t lastmemo;		/* Time of last memo send	*/
    struct u_chanlist {		/* Channels user is on		*/
	struct u_chanlist *next, *prev;
	Channel *chan;		/* Channel Name			*/
    } *chans;
};

/* Channel structure: One per channel */
struct channel_ {
    Channel *next, *prev;
    char name[CHANNELLEN];	/* Channel Name			*/
    char *key;			/* Channel's +k key		*/
#ifdef UNREAL3
    char *link;			/* Channel this one is linked to*/
    char *flood;		/* Channel flood trigger	*/
#endif
    char topic[TOPICLEN];	/* Channel's Topic		*/
    char topicsetter[NICKLEN];	/* Who set the topic		*/
    time_t topictime;		/* When the topic was set	*/
    int32 mode;			/* Channel Modes		*/
    int32 limit;		/* Channel's +l limit		*/
    struct c_userlist {		/* Users on channel		*/
	struct c_userlist *next, *prev;
	User *user;		/* User's NickName		*/
	int32 mode;		/* Mode on channel		*/
    } *users;
};

/* Server information */
typedef struct {
    char *server;		/* Server's name		*/
    char *uplink;		/* Server's uplinke		*/
} Servers;

/* SRAs */
struct sra_ {
    SRA *next, *prev;
    char *nick;			/* SRA's nickname		*/
    char *pass;			/* SRA's password		*/
};

/* Clone Triggers */
struct trigger_ {
    Trigger *next, *prev;
    char host[HOSTLEN];		/* Host for the Trigger		*/
    int limit;			/* Amount this host can have	*/
};

/* Clone Exceptions */
struct exception_ {
    Exception *next, *prev;
    char host[HOSTLEN];		/* Host for the Exception	*/
};

/* AKill structure */
struct akill_ {
    AutoKill *next, *prev;
    char *mask;			/* Mask that's AKilled		*/
    char *reason;		/* Reason for the AKill		*/
    char *setter;		/* Who set the AKill		*/
    int realname;		/* Is it a realname AKill?	*/
    time_t set;			/* When the AKill was set	*/
    time_t expires;		/* When the AKill expires	*/
};

struct timeout_ {
    Timeout *next, *prev;
    time_t settime;		/* When it was set		*/
    uint32 timeout;		/* When it triggers		*/
    int repeat;			/* Does this timeout repeat?	*/
    void (*code)(Timeout *);	/* This structure is passed on	*/
    void *data;			/* Can be anything		*/
};

/* Clone structure */
struct clone_ {
    Clone *next, *prev;
    char host[HOSTLEN];		/* Host for this record		*/
    int cnt;			/* Number of connections	*/
    int kills;			/* Number of kills		*/
    time_t lastkill;		/* When we last killed this host*/
};

/* Verification structure */
struct verify_ {
    Verify *next, *prev;
    char name[64];		/* Object waiting to be verified*/
    time_t since;		/* How long it's been waiting	*/
};

/* List of addresses we've E-Mailed */
struct email_ {
    EMail *next, *prev;
    char *email;		/* The address.			*/
    int times;			/* How many times we've emailed	*/
    time_t since;		/* First time we E-Mailed it	*/
};

/* Channel Access List:
      0: Not in use
      1: VOP
      2: HOP (Unreal only)
      3: AOP
      4: SOP
      5: Founder
 */
typedef struct {
    int32 level;		/* Level of this entry		*/
    char *mask;			/* Mask of entry		*/
} ChanAccess;

/* Channel AKick List */
typedef struct {
    char *mask;			/* Mask of entry		*/
    char *reason;		/* Reason for the AKick		*/
    char setter[NICKLEN];	/* Who set the AKick		*/
    time_t time;		/* When it was set		*/
    int32 level;		/* Level of this entry		*/
} AKick;

/* Memo structure */
struct memo_ {
    int32 number;		/* Memo Number			*/
    int32 flags;		/* Memos flags			*/
    time_t time;		/* When it was sent		*/
    char sender[NICKLEN];	/* Who sent it			*/
    char *text;			/* Actual memo			*/
};

/* Memo Info */
typedef struct {
    int memocnt;		/* Number of memos they have	*/
    int memolimit;		/* Max memos they can have	*/
    Memo *memos;		/* Memo data			*/
} MemoInfo;

/* NickName Information structure */
struct nickinfo_ {
    NickInfo *next, *prev;
    char nick[NICKLEN];		/* Registered NickName		*/
    char pass[PASSWDLEN];	/* Password for NickName	*/
    char *host;			/* Nick this one is linked to	*/
    char *usermask;		/* Last user@host mask		*/
    char *real;			/* Last RealName		*/
    char **access;		/* NickName's Access List	*/
    char **links;		/* List of linked nicks		*/
    char **chans;		/* List of channels owned	*/
    char **successor;		/* List of successor channels	*/
    char *email;		/* E-Mail address		*/
    char *url;			/* URL				*/
    char *name;			/* Real Name			*/
    char *sex;			/* Sex				*/
    char *location;		/* Location			*/
    char *forward;		/* NickName to forward memos to	*/
    char *freezereason;		/* Optional reason for Freeze	*/
    char *temp;			/* For use with temp emails	*/
    unsigned long key;		/* NickServ Key			*/
    int32 age;			/* Age				*/
    int32 memolimit;		/* User Defined MemoLimit	*/
    int32 uin;			/* ICQ UIN			*/
    int32 flags;		/* NickName Flags		*/
    int32 status;		/* NickName Status Flags	*/
    int32 sstamp;		/* Services' Stamp		*/
    int32 accesscnt;		/* Number of Access entries	*/
    int32 linkcnt;		/* Number of Linked Nicks	*/
    int32 chancnt;		/* Number of Channels owned	*/
    int32 successorcnt;		/* Number of Successor channels	*/
    time_t timestamp;		/* TimeStamp			*/
    time_t registered;		/* Time of Registration		*/
    time_t lastseen;		/* Last logoff after ID		*/
    MemoInfo memos;		/* NickName's Memos		*/
};

/* Channel Information structure */
struct chaninfo_ {
    ChanInfo *next, *prev;
    char name[CHANNELLEN];	/* Channel name			*/
    char founder[NICKLEN];	/* Channel founder		*/
    char pass[PASSWDLEN];	/* Channel password		*/
    char topic[TOPICLEN];	/* Channel topic		*/
    char topicsetter[NICKLEN];	/* Who set the topic		*/
    char *successor;		/* Channel successor		*/
    char *greet;		/* Channel Greet message	*/
    char *url;			/* Channel URL			*/
    char *mlock_key;		/* MLock key			*/
#ifdef UNREAL3
    char *mlock_flood;		/* MLock flood			*/
    char *mlock_link;		/* MLock link			*/
#endif
    ChanAccess *access;		/* Access list			*/
    AKick *akick;		/* AutoKick List		*/
    int32 accesscnt;		/* Number of entries in access	*/
    int32 akickcnt;		/* Number of AKicks		*/
    int32 mlock_on;		/* Modes locked on		*/
    int32 mlock_off;		/* Modes locked off		*/
    int32 mlock_limit;		/* MLock limit			*/
    int32 topiclock;		/* TopicLock setting		*/    
    int32 flags;		/* Channel Flags		*/
    time_t topictime;		/* When the topic was set	*/
    time_t registered;		/* Time of Registration		*/
    time_t lastused;		/* Time of last usage		*/
    unsigned long key;		/* ChanServ Key			*/
};

/* Commands we understand: See process.c */
typedef struct {
    const char *name;		/* Name of the command		*/
    void (*func) (char *source, int ac, char **av);
} Message;

extern Message messages[];
extern Message *find_message (const char *name);

/* Command hash */
typedef struct {
    char *accept;		/* Name of the command		*/
    short access;		/* Access needed to use command	*/
    void (*process) (const char *source);
} Hash;

/* Subcommand hash */
typedef struct {
    char *accept;		/* Name of the command		*/
    short access;		/* Access needed to use command	*/
    void (*process) (const char *source);
} Hash_SUB;

/* Help hash */
typedef struct {
    char *accept;		/* Name of the command		*/
    short access;		/* Access needed to use command	*/
    void (*process) (const char *source);
} Hash_HELP;

extern Hash *get_hash (const char *whoami, const char *source, const char *cmd, Hash *hash_table); 
extern Hash_SUB *get_sub_hash (const char *whoami, const char *source, const char *cmd, Hash_SUB *hash_table);
extern Hash_HELP *get_help_hash (const char *whoami, const char *source, const char *cmd, Hash_HELP *hash_table);

/* Hash table access levels */
#define H_NONE  0		/* No access. Everyone can use	*/
#define H_IRCOP 1		/* IRCOp access needed		*/
#define H_CSOP	2		/* CSOp access needed		*/
#define H_SRA   3		/* SRA access needed		*/

/* Channel modes we support */
struct flag {
    char mode;			/* Text version of the mode	*/
    int32 flag;			/* Flag version of the mode	*/
    char prefix;		/* Channel user mode prefix	*/
};

/* Hash table for channel modes */
static const struct flag cmodes[] = {
    {'t',	CMODE_t},
    {'n',	CMODE_n},
    {'s',	CMODE_s},
    {'m',	CMODE_m},
    {'l',	CMODE_l},
    {'i',	CMODE_i},
    {'p',	CMODE_p},
    {'k',	CMODE_k},
#ifndef IRCUP9
    {'r',	CMODE_r},
    {'R',	CMODE_R},
#endif
#if defined (PROMETHEUS) || defined (BAHAMUT) || defined (UNREAL3)
    {'c',	CMODE_c},
    {'O',	CMODE_O},
#endif
#ifdef UNREAL3
    {'Q',	CMODE_Q},
    {'S',	CMODE_S},
    {'K',	CMODE_K},
    {'V',	CMODE_V},
    {'f',	CMODE_f},
    {'H',	CMODE_H},
    {'G',	CMODE_G},
    {'C',	CMODE_C},
    {'u',	CMODE_u},
    {'z',	CMODE_z},
    {'N',	CMODE_N},
    {'L',	CMODE_L},
    {'A',	CMODE_A},
#endif
    {0,		0}
};

static const struct flag cumodes[] = {
    {'o',	CMODE_o,	'@'},
    {'v',	CMODE_v,	'+'},
#ifdef UNREAL3
    {'h',	CMODE_h,	'%'},
    {'a',	CMODE_a,	'~'},
    {'q',	CMODE_q,	'*'},
#endif
    {0,		0,		0}
};

/* Mode stacking structure */
struct modedata
{
    time_t used;		/* When it was first used?		*/
    int last_add;		/* When the last mode was added		*/
    char channel[CHANNELLEN];	/* The channel these modes apply to	*/
    char sender[NICKLEN];	/* The sender (Usually ChanServ)	*/
    int32 binmodes_on;		/* The modes being applied		*/
    int32 binmodes_off;		/* The modes being removed		*/
    char opmodes[MAXMODES*2+1];
    char params[BUFSIZE];	/* Parameters for the modes		*/
    int nparams;		/* Number of parameters			*/
    int paramslen;		/* Length of the parameters		*/
    int dotopic;		/* Should we set the topic, too?	*/
    Timeout *timeout;		/* When these modes will be applied	*/
} modedata[3];

/* This needs to be down here so the stuff in services.h it uses works */
#include "extern.h"
