/* Cygnus database converter

   Cygnus IRC Services - Copyright (c) 2001-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (see the file COPYING); if not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "../inc/dbconvert.h"
#include "../inc/sysconf.h"

static FILE *log_file;
static char opt[256];

NickInfo_Wrecked *nicklists_wrecked[256];
MemoList_Wrecked *memolists_wrecked[256];
ChannelInfo_Wrecked *chanlists_wrecked[256];
SOP_Wrecked *sops_wrecked = NULL;
AutoKill_Wrecked *akill_wrecked = NULL;
Allow_Wrecked *allow_wrecked = NULL;
int nakill_wrecked = 0, akill_size_wrecked = 0;
int nsop_wrecked = 0, sop_size_wrecked = 0;
int nallow_wrecked = 0, allow_size_wrecked = 0;

/* Statistics stuff */
int nin = 0, nout = 0, min = 0, mout = 0;
int cin = 0, cout = 0, ain = 0, aout = 0;
int tin = 0, tout = 0, soin = 0, sout = 0;
int adrop = 0, ndrop = 0, cdrop = 0;
int notes = 0;
struct timeval start;

/* Command-line help message */
void do_help (char *prog)
{
printf("
Syntax: %s options
            
-help                This message
-notes               Display conversion notes
-wrecked             Convert WreckedNet Services DBs

You can stack options, ie, -notes -wrecked, to display notes about
converting WreckedNet databases. Be sure to give the services you
want to work with (in this case, wrecked) last, or the other options
will have no effect.
         
", prog);
}

/* Write to the log */
void log (const char *fmt,...)
{
    va_list args;

    va_start (args, fmt);
 
    if (log_file)
    {
        vfprintf (log_file, fmt, args);
        fputc ('\n', log_file);
        fflush (log_file);
    }
}

/* Case insensitive version of strcmp */
int stricmp (const char *s1, const char *s2)
{
    register int c;

    while ((c = tolower (*s1)) == tolower (*s2))
    {
	if (c == 0)
	    return 0;
	s1++;
	s2++;
    }
    if (c < tolower (*s2))
	return -1; 
    return 1;
}

/* Convert tv to ms */
int tv2ms (struct timeval *tv)
{
    return (tv->tv_sec * 1000) + (int) (tv->tv_usec / 1000);
}

/* smalloc :P */
void *smalloc_wrecked (long size)
{
    void *buf;

    buf = malloc (size);

    if (!buf)
	raise (SIGUSR1);

    return buf;
}

/* srealloc :P */
void *srealloc_wrecked (void *oldptr, long newsize)
{
    void *buf = realloc (oldptr, newsize);

    if (!buf)
        raise (SIGUSR1);

    return buf;
}

/* Convert a mode string to flags */
long string_to_flags (char *modes)
{
    long flags = 0;

    while (*modes) 
    {
	int i;

	for (i = 0; cmodes[i].mode != 0 && cmodes[i].mode != *modes; i++);

	if (!cmodes[i].mode)
	    return 0;

	flags |= cmodes[i].flag;
	*modes++ = 0;
    }

    return flags;
}

/* Get the file version of WreckedNet databases. */
int get_file_version_wrecked (FILE * f, const char *filename)
{
    int version = fgetc (f) << 24 | fgetc (f) << 16 | fgetc (f) << 8 | fgetc (f);

    if (ferror (f) || feof (f))
	return 0;

    else if (version > 6 || version < 1)
	return 0;

    return version;
}

/* Find a nickname for WreckedNet databases */
NickInfo_Wrecked *findnick_wrecked (const char *nick)
{
    NickInfo_Wrecked *ni;

    for (ni = nicklists_wrecked[tolower (*nick)]; ni; ni = ni->next)
	if (stricmp (ni->nick, nick) == 0)
	    return ni;

    return NULL;
}

/* Find a channel for WreckedNet databases */
ChannelInfo_Wrecked *cs_findchan_wrecked (const char *chan)
{   
    ChannelInfo_Wrecked *ci;

    for (ci = chanlists_wrecked[tolower (*chan)]; ci; ci = ci->next)
	if (stricmp (ci->name, chan) == 0)
		return ci;

    return NULL;
}

/* Find a memolist for a nick */
MemoList_Wrecked *find_memolist_wrecked (const char *nick)
{
    MemoList_Wrecked *ml = smalloc_wrecked (sizeof (MemoList_Wrecked));
    int i = -1;
    NickInfo_Wrecked *ni = findnick_wrecked (nick);

    if (ni)
    {
	for (ml = memolists_wrecked[tolower (*(ni->nick))];
	    ml && (i = stricmp (ml->nick, ni->nick)) < 0;
	    ml = ml->next);
    }

    return i == 0 ? ml : NULL;
}

/* Check for duplicate memolists on WreckedNet databases */
MemoList_Wrecked *find_dupelist (const char *nick)
{
    MemoList_Wrecked *ml;
    int i = -1;

    for (ml = memolists_wrecked[tolower (*(nick))];
        ml && (i = stricmp (ml->nick, nick)) < 0;
            ml = ml->next);

    return i == 0 ? ml : NULL;
}

/* Delete a memo list */
void del_memolist_wrecked (MemoList_Wrecked *ml)
{
    if (!ml) return;
    if (ml->next)
	ml->next->prev = ml->prev;
    if (ml->prev)
	ml->prev->next = ml->next;
    else
	memolists_wrecked[tolower (*ml->nick)] = ml->next;
    if (ml->memos)
	free (ml->memos);
    free (ml);
}

/* Delete a channel */
int delchan_wrecked (ChannelInfo_Wrecked *ci)
{
    int i;

    if (ci->next)
	ci->next->prev = ci->prev;
    if (ci->prev)
	ci->prev->next = ci->next;
    else
	chanlists_wrecked[tolower (*ci->name)] = ci->next;
    if (ci->desc)
	free (ci->desc);
    if (ci->url)
	free (ci->url);
    if (ci->mlock_key)
	free (ci->mlock_key);
    if (ci->last_topic)
	free (ci->last_topic);
    for (i = 0; i < ci->accesscount; ++i)
	if (ci->access[i].name)
	    free (ci->access[i].name);
    if (ci->access)
	free (ci->access);
    for (i = 0; i < ci->akickcount; ++i)
    {
	if (ci->akick[i].name)
	    free (ci->akick[i].name);
	if (ci->akick[i].reason)
	    free (ci->akick[i].reason);
    }
    if (ci->akick)
	free (ci->akick);
    free (ci);
    ci = NULL;
    return 1;
}

/* Insert a nickname alphabetically into WreckedNet databases */
void alpha_insert_nick_wrecked (NickInfo_Wrecked *ni)
{
    NickInfo_Wrecked *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists_wrecked[tolower (*nick)];
	ni2 && stricmp (ni2->nick, nick) < 0;
	ni3 = ni2, ni2 = ni2->next);
  
    ni->prev = ni3;
    ni->next = ni2;

    if (!ni3)  
	nicklists_wrecked[tolower (*nick)] = ni;
    else            
	ni3->next = ni;
   
    if (ni2)
	ni2->prev = ni;
}

/* Insert a channel alphabetically into WreckedNet databases */
void alpha_insert_chan_wrecked (ChannelInfo_Wrecked *ci)
{
    ChannelInfo_Wrecked *ci2, *ci3;
    char *chan = ci->name;

    for (ci3 = NULL, ci2 = chanlists_wrecked[tolower (*chan)];
	ci2 && stricmp (ci2->name, chan) < 0;
	ci3 = ci2, ci2 = ci2->next)
	;

    ci->prev = ci3;
    ci->next = ci2;

    if (!ci3)
	chanlists_wrecked[tolower (*chan)] = ci;
    else
	ci3->next = ci;
    if (ci2)
	ci2->prev = ci;
}

/* Insert a memo alphabetically into WreckedNet databases */
void alpha_insert_memolist_wrecked (MemoList_Wrecked *ml)
{
    MemoList_Wrecked *ml2, *ml3;
    char *nick = ml->nick;

    for (ml3 = NULL, ml2 = memolists_wrecked[tolower (*nick)];
	ml2 && stricmp (ml2->nick, nick) < 0;
	ml3 = ml2, ml2 = ml2->next);

    ml->prev = ml3;
    ml->next = ml2;

    if (!ml3)
	memolists_wrecked[tolower (*nick)] = ml;
    else
	ml3->next = ml;
    if (ml2)
	ml2->prev = ml;
}              

/* Read a string from a WreckedNet database */
char *read_string_wrecked (FILE * f, const char *filename)
{
    char *s;
    int len;

    len = fgetc (f) * 256 + fgetc (f);
    s = smalloc_wrecked (len);

    if (len != fread (s, 1, len, f))
	return 0;

    return s;
}

/* Main routine */
int main (int ac, char **av, char **envp)
{
    FILE *f;
    struct rlimit rlim;
    int i;
    char *s;

    /* Core file stuff. */
    if (!getrlimit (RLIMIT_CORE, &rlim))
    {
	rlim.rlim_cur = rlim.rlim_max;
	setrlimit (RLIMIT_CORE, &rlim);
    }

    if (chdir (SERVICES_DIR) < 0)
    {
	perror (SERVICES_DIR);
	return 20;
    }

#if HAVE_UMASK
    umask (077);
#endif

    /* Open the logfile */
    log_file = fopen ("dbconvert.log", "w");

    /* Hi! */
    printf ("Cygnus database converter - Copyright (c) 2001-2002 Darcy Grexton\n");

    /* Parse our command line options */
    for (i = 1; i < ac; ++i)
    {
	s = av[i];

	if (*s == '-' || *s == '/')
	{
	    ++s;
	    if (!stricmp (s, "-help") || !stricmp (s, "?") ||
		  !stricmp (s, "help"))
	    {
		do_help (av[0]);
		return 0;
	    }
	    else if (!stricmp (s, "notes"))
		notes = 1;
	    else if (!stricmp (s, "wrecked"))
		if (!convert_wrecked_db ())
		    return 0;
	}
    }

    i = 0;

    /* If there were no command line options, ask them what they want */
    printf ("\nPlease select the type of services databases you want to convert. If\n");
    printf ("the services you want are not listed below, you cannot convert them\n");
    printf ("at this time. Feel free to contact the author to see if support for\n");
    printf ("them can be added. Read section 4.0.0 of the README for contact info.\n\n");
    printf ("  1) WreckedNet IRC Services v1.2.0\n");
    printf ("  2) You don't have my services!\n\n");
    printf ("Option: ");

    scanf ("%d", &i);

    if (i == 1)
    {
	printf ("\nDo you want to read the WreckedNet database conversion notes?\n\n");
	printf ("[Y/N] ");

	scanf ("%s", &opt);

	if (strchr (opt, 'y') || strchr (opt, 'Y'))
	    notes = 1;

	if (!convert_wrecked_db ())
	    return 0;
    }

    printf ("\nI guess I can't help you.. Sorry.\n\n");

    return 0;
}

/* Convert WreckedNet databases. */
int convert_wrecked_db (void)
{
    int i = 0, problem = 0;
    struct timeval now, tmp;
    char s[256];

    /* Conversion notes? Ok, tell them, then exit. */
    if (notes)
    {
	printf ("\nConversion notes for WreckedNet 1.2.0 -> Cygnus:\n\n");
	printf ("  * Channel AKick setters will all be stored as ChanServ, since\n");
	printf ("    this information is not present in the WreckedNet database.\n\n");
	printf ("  * Channel AKick timestamps will be the time of the conversion,\n");
	printf ("    since this information is not present in the WreckedNet database.\n\n");
	printf ("  * Channel descriptions will be lost, as this is not in Cygnus.\n\n");
	printf ("  * KEEPTOPIC, PRIVATE and JOIN channel flags will be lost, as these\n");
	printf ("    are not in Cygnus.\n\n");
	printf ("  * Channels with topiclock will have topiclock aop after conversion,\n");
	printf ("    as WreckedNet did not have multiple topiclock levels.\n\n");
	printf ("  * Non-nickname channel access list entries will be ignored.\n\n");
	printf ("  * Forbidden channels will be lost in the conversion.\n\n");
	printf ("  * Forbidden nicknames will be lost in the conversion.\n\n");
	printf ("  * Nickname Ignores will be lost, as this is not in Cygnus.\n\n");
	printf ("  * The IRCOP nick flag will be lost, as this is not in Cygnus.\n\n");
	printf ("  * LogonMSGs will not be converted, as Cygnus does not use them.\n\n");
	printf ("Do you want to continue with conversion?\n\n");
	printf ("[Y/N] ");

	scanf ("%s", &opt);

	if (strchr (opt, 'n') || strchr (opt, 'N'))
	{
	    printf ("\nOk.\n\n");
	    return 0;
	}
    }

    printf ("\nPlease enter the path to your current services databases. For example,\n");
    printf ("/home/bob/services\n\n");
    printf ("If you're not sure, type \"exit\" below, go to the directory with your\n");
    printf ("databases, and type pwd. Re run the db converter (\"./dbconvert\") and type\n");
    printf ("what pwd said here.\n\n");
    printf ("Option: ");

    scanf ("%s", &opt);

    if (!stricmp (opt, "exit"))
    {
	printf ("\nOk.\n\n");
	return 0;
    }

    log ("Converting WreckedNet 1.2.0 databases...");

    /* First, tell them what to expect. */
    printf ("\nConverting WreckedNet Services 1.2.0 databases.\n\n");
    printf ("Loading databases... ");

    /* Make note of when we started */
    gettimeofday (&(start), NULL);

    /* Load all of the DBs from disk. If the load function returns anything but 0,
       it's an error.
     */
    if (!(problem = load_wrecked_ns ()))
	printf ("nick.db... ");
    else
    {
	printf ("\n\nProblem loading nick.db! %s.. skipping...\n",
	    problem == 1 ? "Incorrect version." :
	    problem == 2 ? "File not found." :
	    problem == 3 ? "Possible database corruption." : "Unknown problem... ;/");
    }

    if (!(problem = load_wrecked_cs ()))
	printf ("chan.db... ");
    else
    {
	printf ("\n\nProblem loading chan.db! %s.. skipping...\n",
	    problem == 1 ? "Incorrect version." :
	    problem == 2 ? "File not found." :
	    problem == 3 ? "Possible database corruption." : "Unknown problem... ;/");
    }

    if (!(problem = load_wrecked_ms ()))
	printf ("memo.db... ");
    else
    {
	printf ("\n\nProblem loading memo.db! %s.. skipping...\n",
	    problem == 1 ? "Incorrect version." :
	    problem == 2 ? "File not found." :
	    problem == 3 ? "Possible database corruption." : "Unknown problem... ;/");
    }

    if (!(problem = load_wrecked_rs ()))
	printf ("sop.db... akill.db... clone.db... ");
    else
    {
	printf ("\n\nProblem loading one of the OperServ dbs! %s.. skipping...\n",
	    problem == 1 ? "Incorrect version." :
	    problem == 2 ? "File not found." :
	    problem == 3 ? "Possible database corruption." : "Unknown problem... ;/");
    }

    printf ("\nSaving databases... ");

    if (!(problem = save_wrecked_ns ()))
	printf ("nickserv.db... ");
    else
    {
	printf ("\n\nProblem saving nickserv.db! %s\n\n",
	    problem == 1 ? "Can't write to file." :
	    "Unknown problem... ;/");
	return 0;
    }

    if (!(problem = save_wrecked_cs ()))
	printf ("chanserv.db... ");
    else
    {
	printf ("\n\nProblem saving chanserv.db! %s\n\n",
	    problem == 1 ? "Can't write to file." :
	    "Unknown problem... ;/");
	return 0;
    }

    if (!(problem = save_wrecked_rs ()))
	printf ("rootserv.db... ");
    else
    {
	printf ("\n\nProblem saving rootserv.db! %s\n\n",
	    problem == 1 ? "Can't write to file." :
	    "Unknown problem... ;/");
	return 0;
    }

    /* A few nice stats never hurt anyone :) */
    gettimeofday (&now, NULL);
    timersub (&now, &start, &tmp);

    printf ("\n\nConversion statistics:\n");
    printf ("     %d nicknames in, %d nicknames out\n", nin, nout);

    if (nin != nout)
	printf ("     *** Warning: %d nicknames in, but only %d out!\n", nin, nout);

    printf ("     %d channels in, %d channels out\n", cin, cout);

    if (cin != cout)
	printf ("     *** Warning: %d channels in, but only %d out!\n", cin, cout);

    printf ("     %d memos in, %d memos out\n", min, mout);

    if (min != mout)
	printf ("     *** Warning: %d memos in, but only %d out!\n", min, mout);

    printf ("     %d akills in, %d akills out\n", ain, aout);

    if (ain != aout)
	printf ("     *** Warning: %d akills in, but only %d out!\n", ain, aout);

    printf ("     %d triggers in, %d triggers out\n", tin, tout);

    if (tin != tout)
	printf ("     *** Warning: %d triggers in, but only %d out!\n", tin, tout);

    printf ("     %d csops in, %d csops out\n", soin, sout);

    if (soin != sout)
	printf ("     *** Warning: %d csops in, but only %d out!\n", soin, sout);

    if (ndrop)
	printf ("     %d lost nicknames (Probably forbidden, see log)\n", ndrop);

    if (cdrop)
	printf ("     %d lost channels (Probably forbidden, see log)\n", cdrop);

    if (adrop)
	printf ("     %d lost channel access list entries (Probably non-nickname, see log)\n", adrop);

    printf ("     Conversion took %d ms\n", tv2ms (&tmp));

    printf ("\nAll done.\n\n");

    log ("Conversion complete. Stats:");
    log ("%d nicknames in, %d nicknames out", nin, nout);
    log ("%d channels in, %d channels out", cin, cout);
    log ("%d memos in, %d memos out", min, mout);
    log ("%d akills in, %d akills out", ain, aout);
    log ("%d triggers in, %d triggers out", tin, tout);
    log ("%d csops in, %d csops out", soin, sout);

    if (ndrop)
	log ("%d lost nicknames (Probably forbidden)", ndrop);

    if (cdrop)
	log ("%d lost channels (Probably forbidden)", cdrop);

    if (adrop)
	log ("%d lost channel access list entries (Probably non-nickname)", adrop);

    log ("Conversion took %d ms.", tv2ms (&tmp));

    return 0;
}

/* Load WreckedNet 1.2.0 nick.db */
int load_wrecked_ns (void)
{
    FILE *f;
    int i, j, dupenick;
    NickInfo_Wrecked *ni;
    char tmp[264];

    snprintf (tmp, sizeof (tmp), "%s/nick.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

    switch (i = get_file_version_wrecked (f, tmp))
    {
	case 6:
	    for (i = 33; i < 256; ++i)
		while (fgetc (f) == 1)
		{
		    dupenick = 0;
		    ni = smalloc_wrecked (sizeof (NickInfo_Wrecked));

		    if (1 != fread (ni, sizeof (NickInfo_Wrecked), 1, f))
			return 3;

		    ni->flags &= ~(WNI_IDENTIFIED | WNI_RECOGNIZED);

		    if (findnick_wrecked (ni->nick))
			dupenick = 1;

		    alpha_insert_nick_wrecked (ni);
		    nin++;

		    if (ni->email)
		    {
			ni->email = read_string_wrecked (f, tmp);

			if (!strlen (ni->email))
			{
			    free (ni->email);
			    ni->email = NULL;
			}
		    }
		    else
			ni->email = NULL;

		    if (ni->url)
		    {
			ni->url = read_string_wrecked (f, tmp);

			if (!strlen (ni->url))
			{
			    free (ni->url);
			    ni->url = NULL;
			}
		    }
		    else
			ni->url = NULL;

		    ni->last_usermask = read_string_wrecked (f, tmp);
		    ni->last_realname = read_string_wrecked (f, tmp);

		    if (ni->accesscount)
		    {
			char **access;

			access = smalloc_wrecked (sizeof (char *) * ni->accesscount);
			ni->access = access;

			for (j = 0; j < ni->accesscount; ++j, ++access)
			    *access = read_string_wrecked (f, tmp);
		    }

		    if (ni->ignorecount)
		    {
			char **ignore;

			ignore = smalloc_wrecked (sizeof (char *) * ni->ignorecount);
			ni->ignore = ignore;

			for (j = 0; j < ni->ignorecount; ++j, ++ignore)
			    *ignore = read_string_wrecked (f, tmp);
		    }

		    if (dupenick)
		    {
			if (ni->next)
			    ni->next->prev = ni->prev;
			if (ni->prev)
			    ni->prev->next = ni->next;
			else
			    nicklists_wrecked[tolower (*ni->nick)] = ni->next;

			if (ni->email)
			    free (ni->email);
			if (ni->url)
			    free (ni->url);
			if (ni->last_usermask)
			    free (ni->last_usermask);
			if (ni->last_realname)
			    free (ni->last_realname);
			if (ni->access)
			{
			    for (i = 0; i < ni->accesscount; ++i)
				if (ni->access[i])
				    free (ni->access[i]);

			    if (ni->access)
				free (ni->access);
			}

			if (ni->ignore)
			{
			    for (i = 0; i < ni->ignorecount; ++i)
				if (ni->ignore[i])
				    free (ni->ignore[i]);

			    if (ni->ignore)
				free (ni->ignore);
			}

			free (ni);
			ni = NULL;
		    }
		}
	    break;
	case 5:
	case 4:
	case 3:
	case 2:
	case 1:
	default:
	    return 1;
    }

    fclose (f);
    return 0;
}

/* Load a WreckedNet Services 1.2.0 memo.db */
int load_wrecked_ms (void)
{
    FILE *f;
    int i, j, dupelist;
    MemoList_Wrecked *ml;
    Memo_Wrecked *memos;
    char tmp[264];

    snprintf (tmp, sizeof (tmp), "%s/memo.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

    switch (i = get_file_version_wrecked (f, tmp))
    {
	case 6:
	case 5:
	case 4:
	case 3:
	case 2:
	case 1:
	    for (i = 33; i < 256; ++i)
		while (fgetc (f) == 1)
		{
		    dupelist = 0;
		    ml = smalloc_wrecked (sizeof (MemoList_Wrecked));

		    if (1 != fread (ml, sizeof (MemoList_Wrecked), 1, f))
			return 3;

		    if (find_dupelist (ml->nick))
			dupelist = 1;

		    alpha_insert_memolist_wrecked (ml);
		    ml->memos = memos = smalloc_wrecked (sizeof (Memo_Wrecked) * ml->n_memos);
		    fread (memos, sizeof (Memo_Wrecked), ml->n_memos, f);

		    for (j = 0; j < ml->n_memos; ++j, ++memos)
			memos->text = read_string_wrecked (f, tmp);

		    if (!dupelist && findnick_wrecked (ml->nick))
			min = min + ml->n_memos;

		    if (dupelist == 1)
			del_memolist_wrecked (ml);

		    if (!findnick_wrecked (ml->nick))
			del_memolist_wrecked (ml);
		}
	    break;
	default:
	    return 1;
    }

    fclose (f);

    return 0;
}

/* Load a WreckedNet 1.2.0 chan.db */
int load_wrecked_cs (void)
{
    FILE *f;
    int i, j, dupechan;
    ChannelInfo_Wrecked *ci;
    char tmp[264];

    snprintf (tmp, sizeof (tmp), "%s/chan.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

    switch (i = get_file_version_wrecked (f, tmp))
    {
	case 6:
	    for (i = 33; i < 256; ++i)
		while (fgetc (f) == 1)
		{
		    dupechan = 0;
		    ci = smalloc_wrecked (sizeof (ChannelInfo_Wrecked));

		    if (1 != fread (ci, sizeof (ChannelInfo_Wrecked), 1, f))
			return 3;

		    if (cs_findchan_wrecked (ci->name))
			dupechan = 1;

		    alpha_insert_chan_wrecked (ci);

		    ci->desc = read_string_wrecked (f, tmp);

#ifdef HABBER
		    if (ci->greet)
			ci->greet = read_string_wrecked (f, tmp);
#endif

		    if (ci->url)
			ci->url = read_string_wrecked (f, tmp);
		    if (ci->mlock_key)
			ci->mlock_key = read_string_wrecked (f, tmp);
		    if (ci->last_topic)
			ci->last_topic = read_string_wrecked (f, tmp);

		    if (ci->accesscount)
		    {
			ChanAccess_Wrecked *access;
			access = smalloc_wrecked (sizeof (ChanAccess_Wrecked) * ci->accesscount);
			ci->access = access;

			if (ci->accesscount != fread (access, sizeof (ChanAccess_Wrecked),
			    ci->accesscount, f))
			    return 3;

			for (j = 0; j < ci->accesscount; ++j, ++access)
			    access->name = read_string_wrecked (f, tmp);

			j = 0;
			access = ci->access;

			while (j < ci->accesscount)
			{
			    if (access->is_nick < 0)
			    {
				--ci->accesscount;
				free (access->name);
				if (j < ci->accesscount)
				    bcopy (access + 1, access, sizeof (*access) *
					(ci->accesscount - j));
			    }
			    else
			    {
				++j;
				++access;
			    }
			}

			if (ci->accesscount)
			    ci->access = srealloc_wrecked (ci->access,
				sizeof (ChanAccess_Wrecked) * ci->accesscount);
			else
			{
			    free (ci->access);
			    ci->access = NULL;
			}
		    }

		    if (ci->akickcount)
		    {
			AutoKick_Wrecked *akick;
			akick = smalloc_wrecked (sizeof (AutoKick_Wrecked) * ci->akickcount);
			ci->akick = akick;
			if (ci->akickcount != fread (akick, sizeof (AutoKick_Wrecked), ci->akickcount, f))
			    return 3;

			for (j = 0; j < ci->akickcount; ++j, ++akick)
			{
			    akick->name = read_string_wrecked (f, tmp);
			    if (akick->reason)
				akick->reason = read_string_wrecked (f, tmp);
			}

			j = 0;
			akick = ci->akick;

			while (j < ci->akickcount)
			{
			    if (akick->is_nick < 0)
			    {
				--ci->akickcount;
				free (akick->name);
				if (akick->reason)
				    free (akick->reason);
				if (j < ci->akickcount)
				    bcopy (akick + 1, akick, sizeof (*akick) *
					(ci->akickcount - j));
			    }
			    else
			    {
				++j;
				++akick;
			    }
			}

			if (ci->akickcount)
			{
			    ci->akick = srealloc_wrecked (ci->akick,
				sizeof (AutoKick_Wrecked) * ci->akickcount);
			}
			else
			{
			    free (ci->akick);
			    ci->akick = NULL;
			}
		    }

		    if (ci->cmd_access)
		    {
			int n_entries;
			ci->cmd_access = smalloc_wrecked (sizeof (short) * 8);

			n_entries = fgetc (f) << 8 | fgetc (f);

			if (n_entries < 0)
			    return 3;

			if (n_entries <= 8)
			    fread (ci->cmd_access, sizeof (short), n_entries, f);
			else
			{
			    fread (ci->cmd_access, sizeof (short), 8, f);
			    fseek (f, sizeof (short) * (n_entries - 8),
				SEEK_CUR);
			}
		    }

		    if (dupechan == 1)
			delchan_wrecked (ci);
		    else
			cin++;
		}
	    break;
	case 5:
	case 4:
	case 3:
	case 2:
	case 1:
	default:
	    return 3;
    }

    fclose (f);
    return 0;
}

/* Load WreckedNet 1.2.0 sop.db, clone.db and akill.db. */
int load_wrecked_rs (void)
{
    FILE *f;
    NickInfo_Wrecked *ni;
    int i;
    char tmp[264];

    snprintf (tmp, sizeof (tmp), "%s/sop.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

   /* Since we don't HAVE an sop.db in Cygnus, we'll just apply
      the CSOP flag to each of these nicknames as we load them.
    */
   switch (i = get_file_version_wrecked (f, tmp))
   {
	case 6:
	case 5:
	case 4:
	case 3:
	case 2:
	case 1:
	    nsop_wrecked = fgetc (f) * 256 + fgetc (f);

	    if (nsop_wrecked < 8)
		sop_size_wrecked = 16;
	    else
		sop_size_wrecked = 2 * nsop_wrecked;

	    sops_wrecked = srealloc_wrecked (sops_wrecked, sizeof (SOP_Wrecked) * sop_size_wrecked);

	    if (!nsop_wrecked)
		break;

	    soin = nsop_wrecked;

	    if (nsop_wrecked != fread (sops_wrecked, sizeof (SOP_Wrecked), nsop_wrecked, f))
		return 3;

	    break;

	default:
	    return 1;
    }

    fclose (f);

    snprintf (tmp, sizeof (tmp), "%s/akill.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

    switch (i = get_file_version_wrecked (f, tmp))
    {
	case 6:
	case 5:
	case 4:
	case 3:
	case 2:
	    nakill_wrecked = fgetc (f) * 256 + fgetc (f);

	    if (nakill_wrecked < 8)
		akill_size_wrecked = 16;
	    else
		akill_size_wrecked = 2 * nakill_wrecked;

	    akill_wrecked = smalloc_wrecked (sizeof (*akill_wrecked) * akill_size_wrecked);

	    if (!nakill_wrecked || nakill_wrecked == 1)
		break;

	    if (nakill_wrecked != fread (akill_wrecked, sizeof (*akill_wrecked), nakill_wrecked, f))
		return 3;

	    ain = nakill_wrecked;

	    for (i = 0; i < nakill_wrecked; ++i)
	    {
		akill_wrecked[i].mask = read_string_wrecked (f, tmp);
		akill_wrecked[i].reason = read_string_wrecked (f, tmp);
	    }

	    break;

	case 1:
	default:
	    return 1;
    }

    fclose (f);

    snprintf (tmp, sizeof (tmp), "%s/clone.db", opt);

    f = fopen (tmp, "r");

    if (!f)
	return 2;

    switch (i = get_file_version_wrecked (f, tmp))
    {
	case 6:
	case 5:
	case 4:
	case 3:
	case 2:
	case 1:
	    nallow_wrecked = fgetc (f) * 256 + fgetc (f);

	    if (nallow_wrecked < 8)
		allow_size_wrecked = 16;
	    else
		allow_size_wrecked = 2 * nallow_wrecked;

	    allow_wrecked = smalloc_wrecked (sizeof (*allow_wrecked) * allow_size_wrecked);

	    if (!nallow_wrecked || nallow_wrecked == 1)
		break;

	    if (nallow_wrecked != fread (allow_wrecked, sizeof (*allow_wrecked), nallow_wrecked, f))
		return 3;

	    tin = nallow_wrecked;

	    for (i = 0; i < nallow_wrecked; ++i)
	    {
		allow_wrecked[i].host = read_string_wrecked (f, tmp);
		allow_wrecked[i].reason = read_string_wrecked (f, tmp);
	    }

	    break;

	default:
	    return 1;
    }

    fclose (f);

    return 0;
}

/* Save WreckedNet 1.2.0 nick.db as Cygnus nickserv.db */
int save_wrecked_ns ()
{
    FILE *f;
    NickInfo_Wrecked *ni, *tni;
    Memo_Wrecked *m;
    MemoList_Wrecked *ml;
    ChannelInfo_Wrecked *ci;
    int i, j;
    char **access;

    f = fopen ("nickserv.db", "w");

    if (!f)
	return 1;

    for (i = 33; i < 256; ++i)
    {
	for (ni = nicklists_wrecked[i]; ni; ni = ni->next)
	{
	    long outflags = 0;

	    /* We don't do forbids. Skip to the next nick. */
	    if (ni->flags & WNI_VERBOTEN)
	    {
		log ("Skipping forbidden nickname %s", ni->nick);
		ndrop++;
		continue;
	    }

	    /* Check for applicable nickflags */
	    if (ni->flags & WNI_SUSPENDED)
		outflags |= CNF_FROZEN;

	    if (ni->flags & WNI_KILLPROTECT)
		outflags |= CNF_ENFORCE;

	    if (ni->flags & WNI_SECURE)
		outflags |= CNF_SECURE;

	    if (ni->flags & WNI_PRIVATE)
		outflags |= CNF_PRIVATE;

	    if (ni->flags & WNI_PRIVMSG)
		outflags |= CNF_PRIVMSG;

	    if (ni->flags & WNI_SLAVE)
		outflags |= CNF_LINKED;

	    if (ni->flags & WNI_NOMEMO)
		outflags |= CNF_NOMEMO;

	    if (ni->flags & WNI_HIDEMAIL)
		outflags |= CNF_HIDEMAIL;

	    if (ni->flags & WNI_NEVEROP)
		outflags |= CNF_NEVEROP;

	    if (ni->flags & WNI_NOOP)
		outflags |= CNF_NOOP;

	    if (ni->flags & WNI_MARKED)
		outflags |= CNF_MARKED;

	    if (ni->flags & WNI_HOLD)
		outflags |= CNF_HELD;

	    if (ni->flags & WNI_MEMOMAIL)
		outflags |= CNF_MEMOMAIL;

	    /* Set flags for CSOps */
	    for (j = 0; j < nsop_wrecked; ++j)
		if (!stricmp (ni->nick, sops_wrecked[j].nick))
		{
		    outflags |= CNF_CSOP;
		    sout++; 
		}

	    if (!(ni->flags & WNI_SLAVE))
		fprintf (f, "NI %s %s %lu %lu %lu 0 %li 25 0 %s %s\n",
		    ni->nick, ni->pass, ni->time_registered, ni->last_seen,
		    ni->last_signon, outflags, ni->last_usermask, ni->last_realname);

	    nout++;

	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		fprintf (f, "AC %s\n", *access);

	    /* Since we store linked nicks differently, we'll go through every nickname,
	       checking if each one is linked to this one.. Bit cpu intensive, but we're
	       not going to be running that long anyhow so who cares?
	     */
	    for (j = 33; j < 256; ++j)
		for (tni = nicklists_wrecked[j]; tni; tni = tni->next)
		    if (!stricmp (tni->last_usermask, ni->nick))
			fprintf (f, "LN %s\n", tni->nick);

	    /* Go through the channels and make the list for this nick */
	    for (j = 33; j < 256; ++j)
		for (ci = chanlists_wrecked[j]; ci; ci = ci->next)
		    if (!stricmp (ci->founder, ni->nick))
			fprintf (f, "CH %s\n", ci->name);

	    ml = find_memolist_wrecked (ni->nick);

	    if (ml)
	    {
		for (j = 0, m = ml->memos; j < ml->n_memos; ++j, ++m)
		{
		    if (m->flags & WMF_EMAILED)
		    {
			m->flags &= WMF_EMAILED;
			m->flags |= CMF_EMAILED;
		    }

		    fprintf (f, "MO %li %li %lu %s %s\n", m->number, m->flags,
			m->time, m->sender, m->text);

		    mout++;
		}
	    }

	    if (ni->email)
		fprintf (f, "EM %s\n", ni->email);

	    if (ni->url)
		fprintf (f, "UR %s\n", ni->url);

	    if (ni->uin)
		fprintf (f, "UN %li\n", ni->uin);
	}
    }

    fclose (f);

    return 0;
}

/* Save WreckedNet 1.2.0 db as Cygnus db */
int save_wrecked_cs (void)
{
    FILE *f;
    int i, j;
    ChannelInfo_Wrecked *ci;
    ChanAccess_Wrecked *ca;
    AutoKick_Wrecked *ak;

    f = fopen ("chanserv.db", "w");

    if (!f)
	return 1;

    for (i = 33; i < 256; ++i)
    {
	for (ci = chanlists_wrecked[i]; ci; ci = ci->next)
	{
	    long outflags = 0;

	    /* We don't do forbids. Skip to the next channel. */
	    if (ci->flags & WCI_VERBOTEN)
	    {
		log ("Skipping forbidden channel %s", ci->name);
		cdrop++;
		continue;
	    }

	    /* Check for applicable channel flags */
	    if (ci->flags & WCI_VERBOSE)
		outflags |= CCF_VERBOSE;

	    if (ci->flags & WCI_VOPALL)
		outflags |= CCF_VOPALL;

	    /* These are the same thing in Cygnus */
	    if ((ci->flags & WCI_SECURE) || (ci->flags & WCI_SECUREOPS))
		outflags |= CCF_SECURE;

	    if (ci->flags & WCI_RESTRICTED)
		outflags |= CCF_RESTRICTED;

	    if (ci->flags & WCI_SUSPENDED)
		outflags |= CCF_FROZEN;

	    if (ci->flags & WCI_HOLD)
		outflags |= CCF_HELD;

	    if (ci->flags & WCI_MARKED)
		outflags |= CCF_MARKED;

	    fprintf (f, "CI %s %s %s %lu %lu %li %li %li %d 0\n", ci->name,
		ci->founder, ci->founderpass, ci->time_registered, ci->last_used,
		outflags, strlen (ci->mlock_on) ? string_to_flags (ci->mlock_on) : 0,
		strlen (ci->mlock_off) ? string_to_flags (ci->mlock_off) : 0,
		(ci->flags & WCI_TOPICLOCK) ? 3 : 0);

	    /* The founder isn't in the access list, but we need them in there on Cygnus */
	    fprintf (f, "CA %s 5\n", ci->founder);

	    for (ca = ci->access, j = 0; j < ci->accesscount; ++ca, ++j)
	    {
		int newlevel = 0;

		if (ca->is_nick < 0)
		    continue;

		if (!findnick_wrecked (ca->name))
		{
		    log ("Dropping access list entry of %s on %s", ca->name, ci->name);
		    adrop++;
		    continue;
		}

		/* Convert the Wrecked-style db levels to Cygnus-style. */
		if (ca->level <= 9)
		    newlevel = 1;

		if (ca->level > 9 && ca->level <= 14)
		    newlevel = 2;

		if (ca->level > 14 && ca->level <= 24)
		    newlevel = 3;

		if (ca->level > 24 && ca->level <= 49)
		    newlevel = 4;

		if (ca->level >= 50)
		{
		    newlevel = 5;

		    /* No one should have this level.. Warn of it */
		    log ("*** Warning: Founder-level access for %s on %s!", ca->name, ci->name);
		}

		/* If by some miracle we get here and still don't have a level assigned,
		   give them VOP and warn about it in the log..
		 */
		if (!newlevel)
		{
		    newlevel = 1;
		    log ("*** Warning: Couldn't get access level for %s on %s! Giving VOP.", ca->name, ci->name);
		}

		fprintf (f, "CA %s %d\n", ca->name, newlevel);
	    }

	    for (ak = ci->akick, j = 0; j < ci->akickcount; ++ak, ++j)
		fprintf (f, "AK %s ChanServ %d %lu %s\n", ak->name,
		    ak->is_nick ? 2 : 1, time (NULL), ak->reason ? ak->reason : "");

#ifdef HABBER
	    if (ci->greet)
		fprintf (f, "GR %s\n", ci->greet);
#endif

	    if (ci->url)
		fprintf (f, "UR %s\n", ci->url);

	    if (strlen (ci->last_topic))
		fprintf (f, "CT %s %lu %s\n", ci->last_topic_setter ?
		    ci->last_topic_setter : "ChanServ", ci->last_topic_time,
		    ci->last_topic);

	    if (ci->mlock_key)
		fprintf (f, "KY %s\n", ci->mlock_key);

	    if (ci->mlock_limit)
		fprintf (f, "LM %li\n", ci->mlock_limit);

	    cout++;
	}
    }

    fclose (f);

    return 0;
}

/* Save WreckedNet 1.2.0 clone.db and akill.db as Cygnus rootserv.db */
int save_wrecked_rs ()
{
    FILE *f;
    int i;

    f = fopen ("rootserv.db", "w");

    if (!f)
	return 1;

    /* Wrecked doesn't store expire times.. Additionally, there may
       be some clock drift, so we can't do time(NULL)-akill->time..
       So we'll just set it to the Wrecked default of one week.
     */
    if (nakill_wrecked > 1)
	for (i = 0; i < nakill_wrecked; ++i)
	{
	    fprintf (f, "AK %s %s 0 %lu 604800 %s\n", akill_wrecked[i].mask,
		akill_wrecked[i].who, akill_wrecked[i].time, akill_wrecked[i].reason);
	    aout++;
	}

    if (nallow_wrecked > 1)
	for (i = 0; i < nallow_wrecked; ++i)
	{
	    fprintf (f, "TR %s %d\n", allow_wrecked[i].host, allow_wrecked[i].amount);
	    tout++;
	}

    return 0;
}
